from pathlib import Path
from typing import Any

from tbump import bump_files
from tbump.test.conftest import file_contains


def test_bump_files_defaults_to_working_dir(test_repo: Path, monkeypatch: Any) -> None:
    monkeypatch.chdir(test_repo)
    bump_files("1.2.42")

    assert file_contains(test_repo / "package.json", '"version": "1.2.42"')


def test_bump_files_with_repo_path(test_repo: Path) -> None:
    bump_files("1.2.42", test_repo)

    assert file_contains(test_repo / "package.json", '"version": "1.2.42"')



from typing import List, Sequence

import cli_ui as ui

from tbump.action import Action
from tbump.config import ConfigFile
from tbump.file_bumper import FileBumper
from tbump.git_bumper import GitBumper
from tbump.hooks import HooksRunner


class ActionGroup:
    def __init__(
        self,
        dry_run_desc: str,
        desc: str,
        actions: Sequence[Action],
        *,
        should_enumerate: bool = False,
    ):
        self.should_enumerate = should_enumerate
        self.desc = desc
        self.dry_run_desc = dry_run_desc
        self.actions = actions

    def print_group(self, dry_run: bool = False) -> None:
        if not self.actions:
            return
        if dry_run and self.dry_run_desc:
            ui.info_2(self.dry_run_desc)
        if not dry_run and self.desc:
            ui.info_2(self.desc)
        for i, action in enumerate(self.actions):
            if self.should_enumerate:
                ui.info_count(i, len(self.actions), end="")
            action.print_self()

    def execute(self) -> None:
        for action in self.actions:
            action.do()


class UpdateConfig(Action):
    def __init__(self, config_file: ConfigFile, new_version: str):
        self.config_file = config_file
        self.new_version = new_version

    def print_self(self) -> None:
        return

    def do(self) -> None:
        # fmt: off
        ui.info_3(
            "Set current version to", ui.blue, self.new_version, ui.reset,
            "in", ui.bold, self.config_file.path.name
        )
        # fmt: on
        self.config_file.set_new_version(self.new_version)


class Executor:
    def __init__(self, new_version: str, file_bumper: FileBumper):
        self.new_version = new_version
        self.work: List[ActionGroup] = []

        config_file = file_bumper.config_file
        assert config_file
        update_config = UpdateConfig(config_file, new_version)

        update_config_group = ActionGroup(
            f"Would update current version in {config_file.path.name}",
            "Updating current version",
            [update_config],
            should_enumerate=False,
        )
        self.work.append(update_config_group)

        patches = ActionGroup(
            "Would patch these files",
            "Patching files",
            file_bumper.get_patches(new_version),
        )
        self.work.append(patches)

    def add_git_and_hook_actions(
        self, new_version: str, git_bumper: GitBumper, hooks_runner: HooksRunner
    ) -> None:
        before_hooks = ActionGroup(
            "Would run these hooks before commit",
            "Running hooks before commit",
            hooks_runner.get_before_hooks(new_version),
            should_enumerate=True,
        )
        self.work.append(before_hooks)

        git_commands = ActionGroup(
            "Would run these git commands",
            "Performing git operations",
            git_bumper.get_commands(new_version),
        )
        self.work.append(git_commands)

        after_hooks = ActionGroup(
            "Would run these hooks after push",
            "Running hooks after push",
            hooks_runner.get_after_hooks(new_version),
            should_enumerate=True,
        )
        self.work.append(after_hooks)

    def print_self(self, *, dry_run: bool = False) -> None:
        for action_group in self.work:
            action_group.print_group(dry_run=dry_run)

    def run(self) -> None:
        for action_group in self.work:
            action_group.print_group(dry_run=False)
            action_group.execute()



import sys
import textwrap
import urllib.parse
from pathlib import Path
from typing import List, Optional

import attr
import cli_ui as ui
import docopt

from tbump.config import get_config_file
from tbump.error import Error
from tbump.executor import Executor
from tbump.file_bumper import FileBumper
from tbump.git import GitError
from tbump.git_bumper import GitBumper
from tbump.hooks import HooksRunner
from tbump.init import init

TBUMP_VERSION = "6.5.0"

USAGE = textwrap.dedent(
    """
Usage:
  tbump [options] <new_version>
  tbump [options] init [--pyproject] <current_version>
  tbump --help
  tbump --version

Options:
   -h --help          Show this screen.
   -v --version       Show version.
   -C --cwd=<path>    Set working directory to <path>.
   --non-interactive  Never prompt for confirmation. Useful for automated scripts.
   --dry-run          Only display the changes that would be made.
   --only-patch       Only patches files, skipping any git operations or hook commands.
   --no-tag           Do not create a tag
   --no-push          Do not push after creating the commit and/or tag
   --no-tag-push      Create a tag, but don't push it
"""
)


class Canceled(Error):
    def print_error(self) -> None:
        ui.error("Canceled by user")


@attr.s
class BumpOptions:
    working_path: Path = attr.ib()
    new_version: str = attr.ib()
    interactive: bool = attr.ib(default=True)
    dry_run: bool = attr.ib(default=False)


def run(cmd: List[str]) -> None:
    opt_dict = docopt.docopt(USAGE, argv=cmd)
    if opt_dict["--version"]:
        print("tbump", TBUMP_VERSION)
        return

    # when running `tbump init` (with current_version missing),
    # docopt thinks we are running `tbump` with new_version = "init"
    # bail out early in this case
    if opt_dict["<new_version>"] == "init":
        sys.exit(USAGE)

    if opt_dict["--cwd"]:
        working_path = Path(opt_dict["--cwd"])
    else:
        working_path = Path.cwd()

    if opt_dict["init"]:
        current_version = opt_dict["<current_version>"]
        use_pyproject = opt_dict["--pyproject"]
        init(working_path, current_version=current_version, use_pyproject=use_pyproject)
        return

    new_version = opt_dict["<new_version>"]
    bump_options = BumpOptions(working_path=working_path, new_version=new_version)
    if opt_dict["--dry-run"]:
        bump_options.dry_run = True
    if opt_dict["--non-interactive"]:
        bump_options.interactive = False

    operations = ["patch", "hooks", "commit", "tag", "push_commit", "push_tag"]
    if opt_dict["--only-patch"]:
        operations = ["patch"]
    if opt_dict["--no-push"]:
        operations.remove("push_commit")
        operations.remove("push_tag")
    if opt_dict["--no-tag-push"]:
        operations.remove("push_tag")
    if opt_dict["--no-tag"]:
        operations.remove("tag")
        # Also remove push_tag if it's still in the list:
        if "push_tag" in operations:
            operations.remove("push_tag")
    bump(bump_options, operations)


def bump(options: BumpOptions, operations: List[str]) -> None:
    working_path = options.working_path
    new_version = options.new_version
    interactive = options.interactive
    dry_run = options.dry_run

    config_file = get_config_file(options.working_path)
    config = config_file.get_config()

    # fmt: off
    ui.info_1(
        "Bumping from", ui.bold, config.current_version,
        ui.reset, "to", ui.bold, new_version,
    )
    # fmt: on

    git_bumper = GitBumper(working_path, operations)
    git_bumper.set_config(config)
    git_state_error = None
    try:
        git_bumper.check_dirty()
        git_bumper.check_branch_state(new_version)
    except GitError as e:
        if dry_run:
            git_state_error = e
        else:
            raise

    file_bumper = FileBumper(working_path)
    file_bumper.set_config_file(config_file)

    executor = Executor(new_version, file_bumper)

    hooks_runner = HooksRunner(working_path, config.current_version, operations)
    if "hooks" in operations:
        for hook in config.hooks:
            hooks_runner.add_hook(hook)

    executor.add_git_and_hook_actions(new_version, git_bumper, hooks_runner)

    if interactive:
        executor.print_self(dry_run=True)
        if not dry_run:
            proceed = ui.ask_yes_no("Looking good?", default=False)
            if not proceed:
                raise Canceled()

    if dry_run:
        if git_state_error:
            ui.error("Git repository state is invalid")
            git_state_error.print_error()
            sys.exit(1)
        else:
            return

    executor.run()

    if config.github_url and "push_tag" in operations:
        tag_name = git_bumper.get_tag_name(new_version)
        suggest_creating_github_release(config.github_url, tag_name)


def suggest_creating_github_release(github_url: str, tag_name: str) -> None:
    query_string = urllib.parse.urlencode({"tag": tag_name})
    if not github_url.endswith("/"):
        github_url += "/"
    full_url = github_url + "releases/new?" + query_string
    ui.info()
    ui.info("Note: create a new release on GitHub by visiting:")
    ui.info(ui.tabs(1), full_url)


def main(args: Optional[List[str]] = None) -> None:
    # Suppress backtrace if exception derives from Error
    if not args:
        args = sys.argv[1:]
    try:
        run(args)
    except Error as error:
        error.print_error()
        sys.exit(1)
